unit Models;

interface

Uses SysUtils;

Const pi=3.1415926535897932384626433832795;

type

SkPoint=record
  x, y, z:single;
end;

PSkPointArray=^TSkPointArray;
TSkPointArray=array[word] of SkPoint;

PVertArray=^VertArray;
VertArray=array[word] of word;

SkBone=record
  StartPoint:word;
  EndPoint:word;
  numVertices:word;
  Vertarray:PVertArray;
end;

PSkBoneArray=^TSkBoneArray;
TSkBoneArray=array[word] of SkBone;

TVertex=record
  x, y, z:single;
end;

PVertexArray=^TVertexArray;
TVertexArray=array[word] of TVertex;

TFace=array[0..2] of word;

PFaceArray=^TFaceArray;
TFaceArray=array[word] of TFace;

TBoneState=record
  AngleYOZ, AngleXOZ, AngleXOY:single;
end;

PBoneStateArray=^TBoneStateArray;
TBoneStateArray=array[word] of TBoneState;

SkinnedMesh=record
  VertexCount:word;
  FacesCount:word;
  TexVertexCount:word;
  Vertices:PVertexArray;
  Faces:PFaceArray;
  TexVertices:PVertexArray;
  TexFaces:PFaceArray;
  PointCount:word;
  Points:PSkPointArray;
  BoneCount:word;
  Bones:PSkBoneArray;
  renderScale:single;
  DeformationBoneState:PBoneStateArray;
  DeformatedVertices:PVertexArray;
end;

PWordArray=^TWordArray;

BodyPartKeyFrame=record
  BoneCount:word;
  BoneState:PBoneStateArray;
  BoneIndexes:PWordArray;
  KFTimer:integer;
end;

PBodyPartAction=^TBodyPartAction;
TBodyPartAction=record
  Duration:integer;
  KeyFrameCount:word;
  KeyFrames:array[byte] of BodyPartKeyFrame;
end;

TModel=class
  Body:SkinnedMesh;
  BodyParts:array[byte] of array[byte] of PBodypartAction;
  numBodyParts:byte;
  numActions:array[byte] of byte;
  ModelName:ShortString;
  CurrentActions:array[byte] of byte;
  Timers:array[byte] of word;
  Procedure LoadFromFile(filename:string);
  Function TimeLeft:word;
  Procedure PresetRotateAngles(DeltaTimer:word);
  Procedure DrawModel(x1, y1, x2, y2:single); overload;
  Procedure DrawModel(x1, y1, x2, y2, mx1, my1, mx2, my2:single); overload;
  Destructor Destroy; override;
end;

implementation

Uses OpenGL;

Procedure TModel.LoadFromFile(filename:string);
type b4=array[1..2] of word;

Function Getb4ofd(n:single):b4;
var
p1:pointer;
p2:^b4;
begin
p1:=@n;
p2:=p1;
Getb4ofd:=p2^;
end;

Function GetStrOfD(n:single):string;
type
s3=string[3];
var
p1:pointer;
p2:^s3;
begin
p1:=@n;
p2:=p1;
GetStrOfD:=p2^;
end;

var
i,j, m, v:integer;
t:file of single;
n:single;

  Function ReadString:ansistring;
  var
  x:single;
  l, k:integer;
  res:string;
    Function GetLIOfS(n:Single):LongInt;
    var
    p1:pointer;
    p2:^longint;
    begin
      p1:=@n;
      p2:=p1;
      GetLIOfs:=p2^;
    end;
  begin
    res:='';
    read(t, x);
    l:=GetLIOfS(x);
    for k := 1 to l do begin
      read(t, x);
      res:=res+GetStrofD(x);
    end;
    ReadString:=res;
  end;

begin
  assignfile(t, filename);
  reset(T);
  read(t, n);
  if getStrOfD(n)<>'YEM' then
  begin
    closefile(t);
    exit;
  end;
  ModelName:=readstring;
  {Load mesh and skeleton}
  read(t, body.RenderScale);
  read(t, n);
  body.VertexCount:=Getb4ofd(n)[1];
  body.FacesCount:=Getb4ofd(n)[2];
  read(t, n);
  body.TexVertexCount:=Getb4ofd(n)[1];
  GetMem(body.Vertices, body.VertexCount*SizeOf(TVertex));
  GetMem(body.Faces, body.FacesCount*SizeOf(TFace));
  GetMem(body.DeformatedVertices, SizeOf(TVertex)*body.VertexCount);
  GetMem(body.TexVertices, body.TexVertexCount*SizeOf(TVertex));
  GetMem(body.TexFaces, body.FacesCount*Sizeof(TFace));
  for i := 0 to body.VertexCount-1 do begin
    read(t, body.Vertices[i].x);
    read(t, body.Vertices[i].y);
    read(t, body.Vertices[i].z);
  end;
  for i := 0 to body.FacesCount-1 do begin
    read(t, n);
    body.Faces[i][0]:=Getb4ofd(n)[1];
    body.Faces[i][1]:=Getb4ofd(n)[2];
    read(t, n);
    body.Faces[i][2]:=Getb4ofd(n)[1];
  end;
  for i := 0 to body.TexVertexCount-1 do begin
    read(t, body.TexVertices[i].x);
    read(t, body.TexVertices[i].y);
    read(t, body.TexVertices[i].z);
  end;
  for i := 0 to body.FacesCount-1 do begin
    read(t, n);
    body.TexFaces[i][0]:=Getb4ofd(n)[1];
    body.TexFaces[i][1]:=Getb4ofd(n)[2];
    read(t, n);
    body.TexFaces[i][2]:=Getb4ofd(n)[1];
  end;
  read(t, n);
  body.PointCount:=Getb4ofd(n)[1];
  body.BoneCount:=Getb4ofd(n)[2];
  GetMem(body.Points, body.PointCount*sizeof(SkPoint));
  GetMem(body.Bones, body.BoneCount*SizeOf(SkBone));
  for i := 0 to body.PointCount-1 do begin
    read(t, body.Points[i].x);
    read(t, body.Points[i].y);
    read(t, body.Points[i].z);
  end;
  for i := 0 to body.BoneCount-1 do begin
    read(t, n);
    body.Bones[i].StartPoint:=Getb4ofd(n)[1];
    body.Bones[i].EndPoint:=Getb4ofd(n)[2];
    read(t, n);
    body.Bones[i].numVertices:=Getb4ofd(n)[1];
    GetMem(body.Bones[i].VertArray, 2*body.Bones[i].numVertices);
    for j := 0 to body.Bones[i].numVertices-1 do begin
      read(t, n);
      body.Bones[i].VertArray[j]:=Getb4ofd(n)[1];
    end;
  end;
  read(t, n);
  numBodyParts:=Getb4ofd(n)[1];
  for i := 0 to numBodyParts-1 do readString;
  GetMem(body.DeformationBoneState, body.BoneCount*SizeOf(TBoneState));
  for i := 0 to body.BoneCount-1 do begin
    body.DeformationBoneState[i].AngleYOZ :=0;
    body.DeformationBoneState[i].AngleXOZ :=0;
    body.DeformationBoneState[i].AngleXOY :=0;
  end;
  for i := 0 to numBodyParts-1 do begin
    read(t, n);
    numActions[i]:=Getb4ofd(n)[1];
    for j := 0 to numactions[i]-1 do begin
      GetMem(BodyParts[i][j], SizeOf(TBodyPartAction));
      read(t, n);
      BodyParts[i][j].Duration:=Getb4ofd(n)[1];
      BodyParts[i][j].KeyFrameCount:=Getb4ofd(n)[2];
      for m := 0 to BodyParts[i][j].KeyFrameCount-1 do begin
        read(t, n);
        BodyParts[i][j].KeyFrames[m].KFTimer:=Getb4ofd(n)[1];
        BodyParts[i][j].KeyFrames[m].BoneCount:=Getb4ofd(n)[2];
        GetMem(BodyParts[i][j].KeyFrames[m].BoneState, BodyParts[i][j].KeyFrames[m].BoneCount*SizeOf(TBoneState));
        GetMem(BodyParts[i][j].KeyFrames[m].BoneIndexes, BodyParts[i][j].KeyFrames[m].BoneCount*2);
        for v := 0 to Bodyparts[i][j].KeyFrames[m].BoneCount-1 do begin
          read(t, n);
          BodyParts[i][j].KeyFrames[m].BoneState[v].AngleYOZ:=n;
          read(t, n);
          BodyParts[i][j].KeyFrames[m].BoneState[v].AngleXOZ:=n;
          read(t, n);
          BodyParts[i][j].KeyFrames[m].BoneState[v].AngleXOY:=n;
          read(t, n);
          Bodyparts[i][j].KeyFrames[m].BoneIndexes[v]:=GetB4ofd(n)[1];
        end;
      end;
    end;
  end;
  closefile(t);
end;

Function TModel.TimeLeft:word;
var
i:integer;
TimeLeftBP:array[byte] of word;
MinT:word;
begin
  for i := 0 to numBodyParts-1 do begin
    TimeLeftBP[i]:= Bodyparts[i, CurrentActions[i]].Duration - Timers[i];
  end;
  MinT:=65000;
  for i := 0 to numbodyparts-1 do
    if MinT>TimeLeftBP[i] then MinT:=TimeLeftBP[i];
  TimeLeft:=MinT;
end;

Procedure TModel.PresetRotateAngles(DeltaTimer:word);
var
i, j, k:integer;
Anglex1, anglex2, angley1, angley2, anglez1, anglez2, anglex, angley, anglez:single;
minupper, maxlower:integer;
KeyFrameIndex:integer;
begin
  {Increase timers}
  for i := 0 to numBodyParts-1 do begin
    Timers[i]:=Timers[i]+DeltaTimer;
    if Timers[i]>BodyParts[i, CurrentActions[i]].Duration then
      Timers[i]:=Timers[i]-BodyParts[i, currentActions[i]].Duration;
  end;
  keyFrameIndex:=-1;
  for i := 0 to numBodyParts -1 do begin
    {Try to find suitable keyframe}
    KeyFrameIndex:=-1;
    for j := 0 to BodyParts[i, CurrentActions[i]].KeyFrameCount-1 do
      if Timers[i]=BodyParts[i, CurrentActions[i]].KeyFrames[j].KFTimer then
        KeyFrameIndex:=j;
    if keyFrameIndex=-1 then begin
      {Couldn't find suitable keyframes}
      MaxLower:=0;
      MinUpper:=BodyParts[i, CurrentActions[i]].Duration+1;
      for j := 0 to BodyParts[i, CurrentActions[i]].KeyFrameCount-1 do begin
        if BodyParts[i, CurrentActions[i]].KeyFrames[j].KFTimer>Timers[i] then begin
          if Bodyparts[i, CurrentActions[i]].KeyFrames[j].KFTimer<MinUpper then
            MinUpper:=Bodyparts[i, CurrentActions[i]].KeyFrames[j].KFTimer;
        end else
          if BodyParts[i, CurrentActions[i]].KeyFrames[j].KFTimer>MaxLower then
            MaxLower:=BodyParts[i, CurrentActions[i]].KeyFrames[j].KFTimer;
      end;
      for j := 0 to BodyParts[i, CurrentActions[i]].KeyFrames[0].BoneCount-1 do begin
        for k:= 0 to Bodyparts[i, CurrentActions[i]].KeyframeCount-1 do begin
          if MaxLower=BodyParts[i, CurrentActions[i]].KeyFrames[k].KFTimer then begin
            Anglex1:=Bodyparts[i, CurrentActions[i]].keyFrames[k].BoneState[j].AngleYOZ;
            Angley1:=Bodyparts[i, CurrentActions[i]].keyFrames[k].BoneState[j].AngleXOZ;
            Anglez1:=Bodyparts[i, CurrentActions[i]].keyFrames[k].BoneState[j].AngleXOY;
          end;
          if MinUpper=BodyParts[i, CurrentActions[i]].KeyFrames[k].KFTimer then begin
            Anglex2:=Bodyparts[i, CurrentActions[i]].keyFrames[k].BoneState[j].AngleYOZ;
            Angley2:=Bodyparts[i, CurrentActions[i]].keyFrames[k].BoneState[j].AngleXOZ;
            Anglez2:=Bodyparts[i, CurrentActions[i]].keyFrames[k].BoneState[j].AngleXOY;
          end;
        end;
        anglex:=(Anglex2-anglex1)/(minUpper-maxLower)*Timers[i]+anglex1-(Anglex2-anglex1)/(minUpper-maxLower)*maxlower;
        angley:=(Angley2-angley1)/(minUpper-maxLower)*Timers[i]+angley1-(Angley2-angley1)/(minUpper-maxLower)*maxlower;
        anglez:=(Anglez2-anglez1)/(minUpper-maxLower)*Timers[i]+anglez1-(Anglez2-anglez1)/(minUpper-maxLower)*maxlower;
        Body.DeformationBoneState[Bodyparts[i, CurrentActions[i]].KeyFrames[0].BoneIndexes[j]].AngleYOZ:=anglex;
        Body.DeformationBoneState[Bodyparts[i, CurrentActions[i]].KeyFrames[0].BoneIndexes[j]].AngleXOZ:=angley;
        Body.DeformationBoneState[Bodyparts[i, CurrentActions[i]].KeyFrames[0].BoneIndexes[j]].AngleXOY:=anglez;
      end;
    end else begin
      {Found suitable keyframes}
      for j := 0 to BodyParts[i, CurrentActions[i]].KeyFrames[KeyFrameIndex].BoneCount-1 do begin
        Body.DeformationBoneState[Bodyparts[i, CurrentActions[i]].KeyFrames[KeyFrameIndex].BoneIndexes[j]].AngleYOZ:=Bodyparts[i, CurrentActions[i]].KeyFrames[KeyFrameIndex].BoneState[j].AngleYOZ;
        Body.DeformationBoneState[Bodyparts[i, CurrentActions[i]].KeyFrames[KeyFrameIndex].BoneIndexes[j]].AngleXOZ:=Bodyparts[i, CurrentActions[i]].KeyFrames[KeyFrameIndex].BoneState[j].AngleXOZ;
        Body.DeformationBoneState[Bodyparts[i, CurrentActions[i]].KeyFrames[KeyFrameIndex].BoneIndexes[j]].AngleXOY:=Bodyparts[i, CurrentActions[i]].KeyFrames[KeyFrameIndex].BoneState[j].AngleXOY;
      end;
    end;
  end;
end;


Procedure TModel.DrawModel(x1, y1, x2, y2, mx1, my1, mx2, my2:single);

  Procedure RotateBone(BoneIndex:word);
  var
  R_i:integer;
  x, y, z:single;
  x0, y0, z0, d, alpha0:single;
  begin
    for R_i := 0 to Body.BoneCount-1 do
      if Body.Bones[R_i].StartPoint=Body.Bones[BoneIndex].EndPoint then RotateBone(R_i);
    x:=Body.Points[Body.Bones[BoneIndex].StartPoint].x;
    y:=Body.Points[Body.Bones[BoneIndex].StartPoint].y;
    z:=Body.Points[Body.Bones[BoneIndex].StartPoint].z;
    for R_i := 0 to Body.Bones[BoneIndex].numVertices-1 do begin
      x0:=Body.DeformatedVertices[Body.Bones[BoneIndex].VertArray[R_i]].x;
      y0:=Body.DeformatedVertices[Body.Bones[BoneIndex].VertArray[R_i]].y;
      z0:=Body.DeformatedVertices[Body.Bones[BoneIndex].VertArray[R_i]].z;
      d:=sqrt((y0-y)*(y0-y)+(z0-z)*(z0-z));
      alpha0:=arctan((z0-z)/(y0-y));
      if y0-y<0 then begin
        if alpha0<0 then alpha0:=alpha0+pi else alpha0:=alpha0-pi;
      end;
      alpha0:=alpha0+Body.DeformationBoneState[BoneIndex].AngleYOZ;
      y0:=y+d*Cos(alpha0);
      z0:=z+d*sin(alpha0);
      d:=sqrt((x0-x)*(x0-x)+(z0-z)*(z0-z));
      alpha0:=arctan((z0-z)/(x0-x));
      if x0-x<0 then begin
        if alpha0<0 then alpha0:=alpha0+pi else alpha0:=alpha0-pi;
      end;
      alpha0:=alpha0+Body.DeformationBoneState[BoneIndex].AngleXOZ;
      x0:=x+d*Cos(alpha0);
      z0:=z+d*sin(alpha0);
      d:=sqrt((x0-x)*(x0-x)+(y0-y)*(y0-y));
      alpha0:=arctan((y0-y)/(x0-x));
      if x0-x<0 then begin
        if alpha0<0 then alpha0:=alpha0+pi else alpha0:=alpha0-pi;
      end;
      alpha0:=alpha0+Body.DeformationBoneState[BoneIndex].AngleXOY;
      x0:=x+d*Cos(alpha0);
      y0:=y+d*sin(alpha0);
      Body.DeformatedVertices[Body.Bones[BoneIndex].VertArray[R_i]].x:=x0;
      Body.DeformatedVertices[Body.Bones[BoneIndex].VertArray[R_i]].y:=y0;
      Body.DeformatedVertices[Body.Bones[BoneIndex].VertArray[R_i]].z:=z0;
    end;
  end;

var
  PointsDerivation:array[word] of boolean;
  TmpVertex:TVertex;
  TmpFace:TFace;
  TmpTexVertex:TVertex;
  TmpTexFace:TFace;
  i, j, k:integer;

begin
  {Deformate the model}
  for i := 0 to Body.VertexCount-1 do
    Body.DeformatedVertices[i]:=Body.Vertices[i];
  for i := 0 to Body.PointCount-1 do PointsDerivation[i]:=false;
  for i := 0 to Body.PointCount-1 do
    for j := 0 to Body.BoneCount-1 do
      if Body.Bones[j].EndPoint = i then PointsDerivation[i]:=true;
  for i := 0 to Body.BoneCount-1 do
    if not PointsDerivation[Body.Bones[i].StartPoint] then RotateBone(i);
  glEnable(GL_BLEND);
  glDisable(GL_Depth_Test);
  for i := 0 to Body.FacesCount-1 do begin
    TmpFace:=Body.Faces[i];
    TmpTexFace:=Body.TexFaces[i];
    glBlendFunc(GL_DST_COLOR,GL_ZERO);
    glBegin(GL_TRIANGLES);
    for j := 0 to 2 do begin
      TmpVertex:=Body.DeformatedVertices[TmpFace[j]];
      TmpTexVertex:=Body.TexVertices[TmpTexFace[j]];
      glTexCoord2f(TmpTexVertex.x*(mx2-mx1)+mx1, 1-(TmpTexVertex.y*(my2-my1)+my1));
      glVertex3fv(@TmpVertex);
    end;
    glEnd;
    glBlendFunc(gl_ONE, gl_ONE);
    glBegin(GL_Triangles);
    for j := 0 to 2 do begin
      TmpVertex:=Body.DeformatedVertices[TmpFace[j]];
      TmpTexVertex:=Body.TexVertices[TmpTexFace[j]];
      glTexCoord2f(TmpTexVertex.x*(x2-x1)+x1, 1-(TmpTexVertex.y*(y2-y1)+y1));
      glVertex3fv(@TmpVertex);
    end;
  end;
  glEnable(GL_DEPTH_TEST);
  glDisable(GL_BLEND);
end;

{Procedure carries out the skeleton deformation according to the YuraEngine
Model deformation algorythm ((C) Yuriy Ushakov)}
Procedure TModel.DrawModel(x1, y1, x2, y2:single);

  Procedure RotateBone(BoneIndex:word);
  var
  R_i:integer;
  x, y, z:single;
  x0, y0, z0, d, alpha0:single;
  begin
    for R_i := 0 to Body.BoneCount-1 do
      if Body.Bones[R_i].StartPoint=Body.Bones[BoneIndex].EndPoint then RotateBone(R_i);
    x:=Body.Points[Body.Bones[BoneIndex].StartPoint].x;
    y:=Body.Points[Body.Bones[BoneIndex].StartPoint].y;
    z:=Body.Points[Body.Bones[BoneIndex].StartPoint].z;
    for R_i := 0 to Body.Bones[BoneIndex].numVertices-1 do begin
      x0:=Body.DeformatedVertices[Body.Bones[BoneIndex].VertArray[R_i]].x;
      y0:=Body.DeformatedVertices[Body.Bones[BoneIndex].VertArray[R_i]].y;
      z0:=Body.DeformatedVertices[Body.Bones[BoneIndex].VertArray[R_i]].z;
      d:=sqrt((y0-y)*(y0-y)+(z0-z)*(z0-z));
      alpha0:=arctan((z0-z)/(y0-y));
      if y0-y<0 then begin
        if alpha0<0 then alpha0:=alpha0+pi else alpha0:=alpha0-pi;
      end;
      alpha0:=alpha0+Body.DeformationBoneState[BoneIndex].AngleYOZ;
      y0:=y+d*Cos(alpha0);
      z0:=z+d*sin(alpha0);
      d:=sqrt((x0-x)*(x0-x)+(z0-z)*(z0-z));
      alpha0:=arctan((z0-z)/(x0-x));
      if x0-x<0 then begin
        if alpha0<0 then alpha0:=alpha0+pi else alpha0:=alpha0-pi;
      end;
      alpha0:=alpha0+Body.DeformationBoneState[BoneIndex].AngleXOZ;
      x0:=x+d*Cos(alpha0);
      z0:=z+d*sin(alpha0);
      d:=sqrt((x0-x)*(x0-x)+(y0-y)*(y0-y));
      alpha0:=arctan((y0-y)/(x0-x));
      if x0-x<0 then begin
        if alpha0<0 then alpha0:=alpha0+pi else alpha0:=alpha0-pi;
      end;
      alpha0:=alpha0+Body.DeformationBoneState[BoneIndex].AngleXOY;
      x0:=x+d*Cos(alpha0);
      y0:=y+d*sin(alpha0);
      Body.DeformatedVertices[Body.Bones[BoneIndex].VertArray[R_i]].x:=x0;
      Body.DeformatedVertices[Body.Bones[BoneIndex].VertArray[R_i]].y:=y0;
      Body.DeformatedVertices[Body.Bones[BoneIndex].VertArray[R_i]].z:=z0;
    end;
  end;

var
  PointsDerivation:array[word] of boolean;
  TmpVertex:TVertex;
  TmpFace:TFace;
  TmpTexVertex:TVertex;
  TmpTexFace:TFace;
  i, j, k:integer;

begin
  {Deformate the model}
  for i := 0 to Body.VertexCount-1 do
    Body.DeformatedVertices[i]:=Body.Vertices[i];
  for i := 0 to Body.PointCount-1 do PointsDerivation[i]:=false;
  for i := 0 to Body.PointCount-1 do
    for j := 0 to Body.BoneCount-1 do
      if Body.Bones[j].EndPoint = i then PointsDerivation[i]:=true;
  for i := 0 to Body.BoneCount-1 do
    if not PointsDerivation[Body.Bones[i].StartPoint] then RotateBone(i);
  glDisable(GL_BLEND);
  glEnable(GL_Depth_Test);
  for i := 0 to Body.FacesCount-1 do begin
    glBegin(GL_TRIANGLES);
    TmpFace:=Body.Faces[i];
    TmpTexFace:=Body.TexFaces[i];
    for j := 0 to 2 do begin
    TmpVertex:=body.DeformatedVertices[TmpFace[j]];
    TmpTexVertex:=body.TexVertices[TmpTexFace[j]];
    glTexCoord2f(TmpTexVertex.x*(x2-x1)+x1, 1-(TmpTexVertex.y*(y2-y1)+y1));
    glVertex3fv(@TmpVertex);
    end;
    glEnd;
  end;
end;

Destructor TModel.Destroy;
var
  i:integer;
  j:integer;
  k:integer;
begin
  FreeMem(Body.Vertices, SizeOf(TVertex)*Body.VertexCount);
  FreeMem(Body.TexVertices, SizeOf(TVertex)*Body.TexVertexCount);
  FreeMem(Body.Faces, SizeOf(TFace)*Body.FacesCount);
  FreeMem(Body.TexFaces, SizeOf(TFace)*Body.FacesCount);
  FreeMem(Body.Points, SizeOf(SkPoint)*Body.PointCount);
  for i := 0 to Body.BoneCount-1 do
    FreeMem(Body.Bones[i].Vertarray, 2*Body.Bones[i].numVertices);
  FreeMem(Body.Bones, SizeOf(SkBone)*Body.BoneCount);
  for i := 0 to numBodyParts-1 do
    for j := 0 to numActions[i]-1 do begin
      for k := 0 to BodyParts[i, j].KeyFrameCount-1 do begin
        FreeMem(Bodyparts[i, j].KeyFrames[k].BoneState, SizeOf(TBoneState)*BodyParts[i, j].KeyFrames[k].BoneCount);
        FreeMem(Bodyparts[i, j].KeyFrames[k].BoneIndexes, 2*BodyParts[i, j].KeyFrames[k].BoneCount);
      end;
      FreeMem(BodyParts[i, j], SizeOf(TBodyPartAction));
    end;
end;

end.
